#ifndef GLOBPREPT_H
#define GLOBPREPT_H

#include "vector.h"
#include "siftop.h"
#include "list.h"
#include "aliast.h"
#include "globalt.h"
#include "bocont.h"
#include "sourcet.h"
#include "loadelt.h"
#include "dbcrst.h"
#include "dbmatt.h"
#include "entitybocon.h"


#ifndef EXTERN
 #define EXTERN extern
#endif

EXTERN FILE        *Log;     ///< Log file of mechprep
EXTERN siftop      *Top;     ///< Sifel topology
EXTERN dbmatt      *Dbmatt;  ///< database of materials and their parameters
EXTERN dbcrst      *Dbcrst;  ///< database of cross-sections and their parameters

EXTERN gfunct      *Tft;     ///< array of time functions used in time dependent mechanical problems for each load case

// Following arrays have size equal to number of nodes (first index has to be node number)
EXTERN long       **Nod_bocon;   ///< boundary conditions at nodes (prescribed values of unknowns) Nod_bc[i][j] returns id of pvalt for i-th node and j-th lcid
EXTERN list         Nod_bclst;   ///< list of particular prescribed values of uknowns. It is referenced by Nod_bocon
EXTERN long       **Nod_ccn;     ///< prescribed coupled dofs at nodes
EXTERN crsectypet  *Nod_cst;     ///< cross-section types at nodes
EXTERN long        *Nod_csti;    ///< cross-section indeces at nodes
EXTERN long        *Nod_cstdbi;  ///< cross-section type indeces of nodes in cross-section database
EXTERN double     **Nod_inicd;   ///< initial conditions at nodes, Nod_inicd[nod_id][lcid] = value of the initial condition for the given quantity (lcid) and given node (nod_id)
EXTERN long       **Nod_sourcet; ///< nodal sources of transported quantities, Nod_sourcet[nod_id][lcid] = index of quantitiy source in the Nod_srclst
EXTERN list         Nod_srclst;  ///< list of particular nodal sources of quantities, indeces from Nod_sourcet refer to this list

// Following arrays have size equal to number of elements (first index has to be element number)
EXTERN elemtypet     *El_type;   ///< type of elements
EXTERN long          *El_nmat;   ///< number of material types assigned to particular elements
EXTERN mattypet     **El_mattype;///< material type of elements
EXTERN long         **El_matid;  ///< material indeces of elements
EXTERN long         **El_matdbi; ///< material type indeces of elements in material database
EXTERN crsectypet    *El_cst;    ///< cross-section of elements
EXTERN long          *El_csti;   ///< cross-section indeces of elements
EXTERN long          *El_cstdbi; ///< cross-section type indeces of elements in cross-section database
EXTERN long         **El_sourcet;///< sources of transported quantities on elements, Nod_sourcet[elem_id][lcid] = index of quantitiy source in the El_srclst
EXTERN list           El_srclst; ///< list of particular sources of quantities on elements, indeces from El_sourcet refer to this list

EXTERN loadelt     ***El_loadt;   ///< element load, El_load[el_id][lcid] = pointer to loadel structure
EXTERN long         **El_loadtln; ///< lines where the edge|surface|volume El_load has been assigned firstly, El_loadln[el_id][lcid][0|1|2] first line with assignment of edge|surface|volume load
EXTERN long         **El_loadtcol;///< columns where the edge|surface|volume El_load has been assigned firstly, El_loadln[el_id][lcid][0|1|2] first column with assignment of edge|surface|volume load
EXTERN list           El_nv_lst;  ///< list of bnodvalt objects for nodal values whose indeces are used in El_loadt 
EXTERN list           El_trc_lst; ///< list of bnodvalt objects for transmission coefficients whose indeces are used in El_loadt 
EXTERN list           El_trr_lst; ///< list of bnodvalt objects for radiation coefficients whose indeces are used in El_loadt  
EXTERN long          *El_tfunc;   ///< indeces of time fucntions for switching elements on/off

// Auxiliary variables
EXTERN long   Check_unused_prop; ///< Flag for checking of unused properties

#endif
