#include "sourceload.h"
#include "globprep.h"
#include "intools.h"
#include "mathem.h"

sourceload::sourceload()
{
   nlc  = 0L;
   tf   = generalfunct(-1);
   neq  =  0L;
   f    = 0.0;
   eq   = NULL;
   ft   = NULL;
   tabf  = NULL;
   lv   = NULL;
}

sourceload::~sourceload()
{
  for (long i = 0; i < neq; i++)
  {
    delete eq[i];
    delete [] ft[i];
  }
  delete [] eq;
  delete [] ft;
  delete [] lv;
  delete [] tabf;
}

long sourceload::read(FILE *in, long lc)
{
  long i;
  Parser par;

  getlong(in, nlc);
  if ((nlc < 1) || (nlc > lc))
    return(2);
  getlong(in, (long&)tf);
  switch (tf)
  {
    case stat :
      break;
    case pars :
      neq = 1;
      eq = new Equation *[neq];
      ft = new char *[neq];
      ft[0] = new char[256];
      break;
    case tab  :
      tabf = new tablefunct;
      getlong(in, i);
      tabf->asize = i;
      tabf->x = new double[i];
      tabf->y = new double[i];
      getlong(in, i);
      tabf->itype = interpoltype(i);
      break;
    case pars_set :
      getlong(in, neq);
      eq = new Equation *[neq];
      ft = new char *[neq];
      lv = new double[neq];
      for (i = 0; i < neq; i++)
        ft[i] = new char[256];
      break;
    default :
      fprintf(stderr, "\n\n Unknown type of time function is required in function sourceload::read().\n");
      fprintf(stderr, "(file %s, line %d)\n", __FILE__, __LINE__);
      return(2);
  }
  switch (tf)
  {
    case stat :
      getdouble(in, f);
      break;
    case pars :
      getstring(in, ft[0], 255);
      eq[0] = par.TextToTree(ft[0]);
      if (eq[0] == NULL)
      {
        fprintf(stderr, "\n\n Unable to parse function sourceload::read().\n");
        fprintf(stderr, "(file %s, line %d)\n", __FILE__, __LINE__);
        return (3);
      }
      break;
    case tab  :
      for (i = 0; i < tabf->asize; i++)
      {
        getdouble(in, tabf->x[i]);
        getdouble(in, tabf->y[i]);
      }
      break;
    case pars_set :
      for (i = 0; i < neq; i++)
      {
        getdouble(in, lv[i]);
        getstring(in, ft[i], 255);
        eq[i] = par.TextToTree(ft[i]);
        if (eq[i] == NULL)
        {
          fprintf(stderr, "\n\n Unable to parse function sourceload::read().\n");
          fprintf(stderr, "(file %s, line %d)\n", __FILE__, __LINE__);
          return (3);
        }
      }
      break;
    default :
      fprintf(stderr, "\n\n Unknown type of time function is required in function sourceload::read().\n");
      fprintf(stderr, "(file %s, line %d)\n", __FILE__, __LINE__);
      return (2);
  }
  return(0);
}
