#include <stdlib.h>
#include "globmatt.h"
#include "globalt.h"
#include "lhsrhst.h"
#include "constrel.h"
#include "onemedium.h"
#include "twomedia.h"
#include "threemedia.h"
#include "fourmedia.h"
#include "elemswitcht.h"

/**
   function assembles conductivity %matrix of the problem
   
   @param lcid - load case id
*/
void conductivity_matrix (long lcid)
{
  long i,ndofe,*cn;
  matrix lm;
  
  if (Kmat==NULL)  Kmat = new gmatrix;
  //Kmat->ts=(storagetype)(*(int*)&Tp->tstorkm);
  Kmat->setval (Tp->ssle);
  Kmat->initiate (Gtt,Ndoft,Tp->tstorkm,Mesprt,Outt);
  
  //  loop over elements
  for (i=0;i<Tt->ne;i++){
    if (Gtt->leso[i]==1){
      //  number of DOFs on element
      ndofe = Gtt->give_ndofe (i);
      
      allocm (ndofe,ndofe,lm);
      //  conductivity matrix of one element
      conductmat (i,lcid,lm);
      
      cn = new long [ndofe];
      Gtt->give_code_numbers (i,cn);
      
      Kmat->localize (lm,cn,i);
      
      destrm (lm);
      delete [] cn;
    }
  }
  
  Kmat->prepmat (0.0,Mesprt);

}

/**
   function assembles capacity %matrix of the problem

   @param lcid - load case id
*/
void capacity_matrix (long lcid)
{
  long i,ndofe,*cn;
  matrix lm;
  
  if (Cmat==NULL)  Cmat = new gmatrix;
  //Cmat->ts=(storagetype)(*(int*)&Tp->tstorcm);
  Cmat->setval (Tp->ssle);
  Cmat->initiate (Gtt,Ndoft,Tp->tstorcm,Mesprt,Outt);

  //  loop over elements
  for (i=0;i<Tt->ne;i++){
    if (Gtt->leso[i]==1){
      //  number of DOFs on element
      ndofe = Gtt->give_ndofe (i);
      
      allocm (ndofe,ndofe,lm);
      //  capacity matrix of one element
      capacmat (i,lcid,lm);
      
      cn = new long [ndofe];
      Gtt->give_code_numbers (i,cn);
      
      Cmat->localize (lm,cn,i);
      
      destrm (lm);
      delete [] cn;
    }
  }
  
  Cmat->prepmat (0.0,Mesprt);
  
}

/**
   @param r - %vector of residuum
   @param p - %vector d_{n+1} - d_{n}
   @param v - auxiliary %vector
   @param dt - actual time increment
   @param lcid - load case id
   
   JK, 2.2.2003
*/
void residuum (double *r,double *p,double *v,double dt,long n,long lcid)
{
  capacity_matrix (0);
  conductivity_matrix (0);
  trfel_right_hand_side (0,Lsrst->rhs,Ndoft);
  
  Cmat->gmxv (p,v);
  cmulv (1.0/dt,v,n);
  Kmat->gmxv (Lsrst->lhs,r);
  addv (r,v,n);
  subv (r,Lsrst->rhs,n);
}


/**
   function computes and extracts values on one element of PUC

   @param lcid - number of load case
   @param r - vector of nodal values

   TKr, 08/09/2010
*/
void elemvalues_puc (long lcid,long eid,double *r,long ndofe)
{
  long i,j,ii,nn,ncomp,ndofn,nne,k;
  ivector nod;
  double *rd;
  j = 0;
  nn = 0; //element node number

  ncomp = Tt->give_ncomp(eid);
  //  number of nodes on element
  nne = Tt->give_nne (eid);
  allocv (nne,nod);
  Tt->give_elemnodes (eid,nod);

  //updating of unknown from master unknown, gradients and fluctuation part
  ii = 0;
  for (i=0;i<nne;i++){
    k = nod[i];
    ncomp = Tt->nodes[k].ncompgrad;
    ndofn=Gtt->give_ndofn (k);
    rd = new double [ndofn];
    nodval (0, rd, k);
    for (j=0;j<ndofn;j++){    
      if(ncomp == 1){
	r[ii] = rd[j] + Tb->lc[j].masterval + Tb->lc[j].mastergrad[0]*Gtt->gnodes[k].x;
      }
      if(ncomp == 2){
	r[ii] = rd[j] + Tb->lc[j].masterval + Tb->lc[j].mastergrad[0]*Gtt->gnodes[k].x+Tb->lc[j].mastergrad[1]*Gtt->gnodes[k].y;
      }
      if(ncomp == 3){
	r[ii] = rd[j] + Tb->lc[j].masterval + Tb->lc[j].mastergrad[0]*Gtt->gnodes[k].x+Tb->lc[j].mastergrad[1]*Gtt->gnodes[k].y+Tb->lc[j].mastergrad[2]*Gtt->gnodes[k].z;	
      }
      ii++;
    }
    delete [] rd;
  }
}


/**
   function extracts values on one element

   @param lcid - number of load case
   @param r - vector of nodal values
   @param cn - array containing code numbers
   @param ndofe - number of DOFs on actual element

   9.7.2001
*/
void elemvalues (long lcid,long eid,double *r,long *cn,long ndofe)
{
  long i,j,ii,nn,ncomp,no,nne;
  ivector nod;
  j = 0;
  nn = 0; //element node number

  ncomp = Tt->give_ncomp(eid);
  //  number of nodes on element
  nne = Tt->give_nne (eid);
  allocv (nne,nod);
  Tt->give_elemnodes (eid,nod);

  switch (Tp->tprob){
  case stationary_problem:{
    for (i=0;i<ndofe;i++){
      ii=cn[i];
      
      //only for homogenization
      //search index of unknown and element node
      if(Tp->homogt == 1){
	j++;
	if(j > Tp->ntm){
	  j = 1;//index of unknown
	  nn++;//node number
	}
      }
      
      if (ii<0){
	if(Tp->homogt == 1){
	  r[i]=Tb->lc[lcid].pv[0-ii-1].getval() + Tb->lc[j-1].masterval;
	  no = nod[nn];
	  if(ncomp == 1){
	    r[i]=r[i]+Tb->lc[j-1].mastergrad[0]*Gtt->gnodes[no].x;
	  }
	  if(ncomp == 2){
	    r[i]=r[i]+Tb->lc[j-1].mastergrad[0]*Gtt->gnodes[no].x+Tb->lc[j-1].mastergrad[1]*Gtt->gnodes[no].y;
	  }
	  if(ncomp == 3){
	    r[i]=r[i]+Tb->lc[j-1].mastergrad[0]*Gtt->gnodes[no].x+Tb->lc[j-1].mastergrad[1]*Gtt->gnodes[no].y+Tb->lc[j-1].mastergrad[2]*Gtt->gnodes[no].z;	
	  }
	}
	else
	  r[i]=Tb->lc[lcid].pv[0-ii-1].getval();
      }
      
      if (ii==0)  r[i]=0.0;
      if (ii>0)   r[i]=Lsrst->lhs[ii-1];
    }
    break;
  }
  case nonlinear_stationary_problem:
  case nonstationary_problem:
  case nonlinear_nonstationary_problem:
  case discont_nonstat_problem:
  case discont_nonlin_nonstat_problem:
  case growing_np_problem:
  case growing_np_problem_nonlin:
  case parallel_homog:{
    for (i=0;i<ndofe;i++){
      ii=cn[i];
      if (ii<0)   r[i]=Tb->lc[0].pv[0-ii-1].getval();
      if (ii==0)  r[i]=0.0;
      if (ii>0){   
	r[i]=Lsrst->lhsi[ii-1]+Lsrst->lhs[ii-1];
      }
    }
    break;
  }
  default:{
    print_err("unknown problem type is required",__FILE__,__LINE__,__func__);
  }
  }

  destrv (nod);
}



/**
   function extracts values on one node
   
   @param lcid  - number of load case
   @param r     - allocated array for nodal values
   @param idn   - node number

   9.7.2001
*/
void nodval (long lcid,double *r, long idn)
{
  long i,ii;
  
  switch (Tp->tprob){
  case stationary_problem:{
    for (i=0;i<Tt->give_ndofn (idn);i++){
      ii=Tt->give_dof(idn,i);
      if (ii<0){ r[i]=Tb->lc[lcid].pv[0-ii-1].getval();}
      if (ii==0)  r[i]=0.0;
      if (ii>0)   r[i]=Lsrst->lhs[ii-1];
    }
    break;
  }
  case nonlinear_stationary_problem:
  case nonstationary_problem:{
    for (i=0;i<Tt->give_ndofn (idn);i++){
      ii=Tt->give_dof(idn,i);
      if (ii<0)   r[i]=Tb->lc[0].pv[0-ii-1].getval();
      if (ii==0)  r[i]=0.0;
      if (ii>0)      if (ii>0){   
	r[i]=Lsrst->lhsi[ii-1]+Lsrst->lhs[ii-1];
      }
    }
    
    break;
  }
  case nonlinear_nonstationary_problem:{
    for (i=0;i<Tt->give_ndofn (idn);i++){
      ii=Tt->give_dof(idn,i);
      if (ii<0)   r[i]=Tb->lc[0].pv[0-ii-1].getval();
      if (ii==0)  r[i]=0.0;
      if (ii>0)      if (ii>0){   
	r[i]=Lsrst->lhsi[ii-1]+Lsrst->lhs[ii-1];
      }
    }
    
    break;
  }
  case discont_nonstat_problem:
  case discont_nonlin_nonstat_problem:{
    for (i=0;i<Tt->give_ndofn (idn);i++){
      ii=Tt->give_dof(idn,i);
      if (ii<0)   r[i]=Tb->lc[0].pv[0-ii-1].getval();
      if (ii==0)  r[i]=0.0;
      if (ii>0){   
	r[i]=Lsrst->lhsi[ii-1]+Lsrst->lhs[ii-1];
      }
    }
    
    break;
  }
    
  case growing_np_problem:
  case growing_np_problem_nonlin:{
    for (i=0;i<Tt->give_ndofn (idn);i++){
      ii=Tt->give_dof(idn,i);
      if (ii<0)   r[i]=Tb->lc[0].pv[0-ii-1].getval();
      if (ii==0)  r[i]=0.0;
      if (ii>0){   
	r[i]=Lsrst->lhsi[ii-1]+Lsrst->lhs[ii-1];
      }
    }
    
    //  subtraction of initial nodal values
    //Tt->elements[eid].subtrinitnodval (r,Tt->give_ndofn (idn));
    break;
  }
  default:{
    print_err("unknown problem type is required",__FILE__,__LINE__,__func__);
  }
  }
}

/**
   function extracts initial values on one node
   
   @param lcid  - number of load case
   @param r     - allocated array for nodal values
   @param idn   - node number

   JK, 18.8.2008
*/
void initnodval (long lcid,double *r, long idn)
{
  long i,ii;
  
  switch (Tp->tprob){
  case nonlinear_stationary_problem:
  case nonstationary_problem:{
    for (i=0;i<Tt->give_ndofn (idn);i++){
      ii=Tt->give_dof(idn,i);
      if (ii<0)   r[i]=Tb->lc[0].pv[0-ii-1].getval();
      if (ii==0)  r[i]=0.0;
      if (ii>0){   
	r[i]=Lsrst->lhsi[ii-1];
      }
    }
    
    break;
  }
  case nonlinear_nonstationary_problem:{
    for (i=0;i<Tt->give_ndofn (idn);i++){
      ii=Tt->give_dof(idn,i);
      if (ii<0)   r[i]=Tb->lc[0].pv[0-ii-1].getval();
      if (ii==0)  r[i]=0.0;
      if (ii>0){   
	r[i]=Lsrst->lhsi[ii-1];
      }
    }
    
    break;
  }
  case discont_nonstat_problem:
  case discont_nonlin_nonstat_problem:{
    for (i=0;i<Tt->give_ndofn (idn);i++){
      ii=Tt->give_dof(idn,i);
      if (ii<0)   r[i]=Tb->lc[0].pv[0-ii-1].getval();
      if (ii==0)  r[i]=0.0;
      if (ii>0){   
	r[i]=Lsrst->lhsi[ii-1];
      }
    }
    
    break;
  }
    
  case growing_np_problem:
  case growing_np_problem_nonlin:{
    for (i=0;i<Tt->give_ndofn (idn);i++){
      ii=Tt->give_dof(idn,i);
      if (ii<0)   r[i]=Tb->lc[0].pv[0-ii-1].getval();
      if (ii==0)  r[i]=0.0;
      if (ii>0){   
	r[i]=Lsrst->lhsi[ii-1];
      }
    }
    
    break;
  }
  default:{
    print_err("unknown problem type is required",__FILE__,__LINE__,__func__);
  }
  }
}


/**
   function extracts values from previous step on one node
   this function differs from the function nodval
   the %vector of values is an input parameter while the
   function nodval uses the %vector Lsrst->lhs
   
   @param lcid - number of load case
   @param r - allocated array for displacement
   @param idn - node number
   @param lhs - %vector of values from the previous step
   
   9.7.2001
*/
void prevnodval (long lcid,double *r, long idn,double *lhs)
{
  long i,ii;
  
  switch (Tp->tprob){
  case stationary_problem:{
    for (i=0;i<Tt->give_ndofn (idn);i++){
      ii=Tt->give_dof(idn,i);
      if (ii<0)   r[i]=Tb->lc[lcid].pv[0-ii-1].getval();
      if (ii==0)  r[i]=0.0;
      if (ii>0)   r[i]=lhs[ii-1];
    }
    break;
  }
  case nonlinear_stationary_problem:
  case nonstationary_problem:{
    for (i=0;i<Tt->give_ndofn (idn);i++){
      ii=Tt->give_dof(idn,i);
      if (ii<0)   r[i]=Tb->lc[0].pv[0-ii-1].getval();
      if (ii==0)  r[i]=0.0;
      if (ii>0){   
	r[i]=Lsrst->lhsi[ii-1]+lhs[ii-1];
      }
    }
    
    break;
  }
  case nonlinear_nonstationary_problem:{
    for (i=0;i<Tt->give_ndofn (idn);i++){
      ii=Tt->give_dof(idn,i);
      if (ii<0)   r[i]=Tb->lc[0].pv[0-ii-1].getval();
      if (ii==0)  r[i]=0.0;
      if (ii>0){
	r[i]=Lsrst->lhsi[ii-1]+lhs[ii-1];
      }
    }
    
    break;
  }
  case discont_nonstat_problem:
  case discont_nonlin_nonstat_problem:{
    for (i=0;i<Tt->give_ndofn (idn);i++){
      ii=Tt->give_dof(idn,i);
      if (ii<0)   r[i]=Tb->lc[0].pv[0-ii-1].getval();
      if (ii==0)  r[i]=0.0;
      if (ii>0){   
	r[i]=Lsrst->lhsi[ii-1]+lhs[ii-1];
      }
    }
    
    break;
  }
    
  case growing_np_problem:
  case growing_np_problem_nonlin:{
    for (i=0;i<Tt->give_ndofn (idn);i++){
      ii=Tt->give_dof(idn,i);
      if (ii<0)   r[i]=Tb->lc[0].pv[0-ii-1].getval();
      if (ii==0)  r[i]=0.0;
      if (ii>0){   
	r[i]=Lsrst->lhsi[ii-1]+lhs[ii-1];
      }
    }
    
    //  subtraction of initial nodal values
    //Tt->elements[eid].subtrinitnodval (r,Tt->give_ndofn (idn));
    break;
  }
  default:{
    print_err("unknown problem type is required",__FILE__,__LINE__,__func__);
  }
  }
}



/**
   function extracts nodal values stored in array other
   values are extracted for all nodes of required element
   
   @param r - array containing nodal values
   @param nod - list of nodes on element

   17.9.2005, JK
*/
void nodalotherval (double *r,ivector &nod)
{
  long i,j,k,ncompo;
  
  k=0;
  for (i=0;i<nod.n;i++){
    ncompo = Tt->nodes[nod[i]].ncompother;
    for (j=0;j<ncompo;j++){
      r[k]=Tt->nodes[nod[i]].other[j];
      k++;
    }
  }

}

/**
   function extracts first time derivatives of nodal values on one element

   @param r - vector of derivatives of nodal values
   @param cn - array containing code numbers
   @param ndofe - number of DOFs on actual element

   4.3.2003
*/
void nodalderivatives (double *r,long *cn,long ndofe)
{
  long i,ii;
  
  switch (Tp->tprob){
  case nonstationary_problem:
  case nonlinear_nonstationary_problem:
  case growing_np_problem:
  case discont_nonstat_problem:
  case discont_nonlin_nonstat_problem:
  case growing_np_problem_nonlin:{
    for (i=0;i<ndofe;i++){
      ii=cn[i];
      if (ii<0)   r[i]=0.0;
      if (ii==0)  r[i]=0.0;
      if (ii>0)   r[i]=Lsrst->tdlhs[ii-1];
    }
    break;
  }
  default:{
    print_err("unknown problem type is required",__FILE__,__LINE__,__func__);
  }
  }
}

/**
   function extracts prescribed and initial values on one element
   
   @param r - %vector of nodal values
   @param cn - array containing code numbers
   @param ndofe - number of DOFs in element
   
   21.3.2004, JK
*/
void prescvalues (double *r,long *cn,long ndofe)
{
  long i,ii;
  
  for (i=0;i<ndofe;i++){
    ii=cn[i];
    if (ii<0)   r[i]=Tb->lc[0].pv[0-ii-1].getval();
    if (ii==0)  r[i]=0.0;
    if (ii>0)   r[i]=Lsrst->lhsi[ii-1];
  }
}

/**
   function extracts initial values on one element
   
   @param r - %vector of nodal values
   @param cn - array containing code numbers
   @param ndofe - number of DOFs on element
   
   21.3.2004, JK
*/
void initialvalues (double *r,long *cn,long ndofe)
{
  long i,ii;
  
  for (i=0;i<ndofe;i++){
    ii=cn[i];
    if (ii<0)   r[i]=Tb->lc[0].pv[0-ii-1].getipv();
    if (ii==0)  r[i]=0.0;
    if (ii>0)   r[i]=Lsrst->lhsi[ii-1];
  }
}

/**
   function computes quantity values in integration points of PUC
   values in integration points are obtained from nodal values of PUC
   
   TKr, 08/09/2010
*/
void approximation_puc ()
{
  long i;
  
  //  loop over elements
  for (i=0;i<Tt->ne;i++){
    if (Gtt->leso[i]==1){
      //  only elements switched on are processed
      
      //  interpolation of nodal values to integration points
      intpointvalues_puc (i);
      //  computation of gradients in integration points
      //intpointgradients (i);
    }
  } 
  
  Tm->aux_values (0);
}


/**
   function computes quantity values in integration points
   values in integration points are obtained from nodal values
   
   16.1.2002
*/
void approximation ()
{
  long i,ii,j,ipp,nip;
  
  if (Tp->nvs==1){
    //  copy of nodal values to array nodval on nodes
    //  implementation due to saltmat4
    copy_nodval (0);
  }
  
  //  loop over elements
  for (i=0;i<Tt->ne;i++){
    if (Gtt->leso[i]==1){
      //  only elements switched on are processed
      
      //  interpolation of nodal values to integration points
      intpointvalues (i);
      //  computation of gradients in integration points
      intpointgradients (i);
    }
  }
  
  
  Tm->aux_values (0);
  
  if (Tp->nvs==1){
    //  arrays of actual values are moved to arrays of previous values
    //  implementation due to saltmat4
    actual_previous_change ();
  }
  
  
  //pro V. Smilauera prumerovani teplot na prvku a ukladani do pole other na int. bodech 
  //provizorni umisteni ??!!
  if (Tm->cemhydr){
    //nulovani prumernych teplot na materialu
    //nulovani poctu int. bodu materialu
    for (i=0;i<Tm->nmt;i++){
      if (Tm->mattype[i] == cementhydrmat)
	for (j=0;j<Tm->numtype[i];j++){
	  Tm->cemhydr[j].average_temp = 0.0;
	  Tm->cemhydr[j].n_ipp = 0;
	}
    }

    //ukladani teplot do materialu
    //ukladani poctu int. bodu materialu
    for (i=0;i<Tt->ne;i++){
      if (Gtt->leso[i]==1){
	ipp=Tt->elements[i].ipp[0][0];
	nip=Tt->give_tnip (i);
	for (j=0;j<nip;j++){
	  ii = Tm->ip[ipp+j].idm;

	  if(Tp->tmatt == onemedium)
	    Tm->cemhydr[ii].average_temp += Tm->ip[ipp+j].av[0];
	  if(Tp->tmatt == twomediacoup)
	    Tm->cemhydr[ii].average_temp += Tm->ip[ipp+j].av[1];
	  if(Tp->tmatt == threemediacoup)
	    Tm->cemhydr[ii].average_temp += Tm->ip[ipp+j].av[2];
	  
	  Tm->cemhydr[ii].n_ipp++;	  
	}
      }
    }
    
    //prumerovani teplot na materialu
    for (i=0;i<Tm->nmt;i++){
      if (Tm->mattype[i] == cementhydrmat)
	for (j=0;j<Tm->numtype[i];j++){
	  if (Tm->cemhydr[j].n_ipp != 0)
	    Tm->cemhydr[j].average_temp /= Tm->cemhydr[j].n_ipp;
	}
    }

    //ukladani prumernych teplot do materialu
    for (i=0;i<Tt->ne;i++){
      if (Gtt->leso[i]==1){
	ipp=Tt->elements[i].ipp[0][0];
	nip=Tt->give_tnip (i);
	for (j=0;j<nip;j++){
	  ii = Tm->ip[ipp+j].idm;
	  Tm->ip[ipp+j].other[0] = Tm->cemhydr[ii].average_temp;
	}
      }
    }
  }
}



/**
   function moves arrays with actual values to arrays of previous values at all integration points
   
   JK, 29.5.2007
*/
void actual_previous_change ()
{
  long i;
  
  for (i=0;i<Tm->tnip;i++){
    Tm->ip[i].actual_previous_change ();
  }
}

/**
   function moves arrays with actual values to arrays of previous values at all nodes
   
   JK, 14.1.2008
*/
void actual_previous_nodval ()
{
  long i;
  
  for (i=0;i<Tt->nn;i++){
    Tt->nodes[i].actual_previous_change ();
  }
}

/**
   function assembles part of the %vector of right hand side
   
   @param rhs - right hand side
   
   21.3.2004, JK
*/
void assemble_init (double *rhs)
{
  long i,ndofe,*cn,lcid=0;
  vector r,f;
  matrix km;
  long ne = Tt->ne;
  
  for (i=0;i<ne;i++){
    
    if (Gtt->leso[i]==1){
      //  only elements switched on are processed
      
      //  number of DOFs on element
      ndofe=Tt->give_ndofe(i);
      allocm (ndofe,ndofe,km);
      
      //  computation of element conductivity matrix
      conductmat (i,lcid,km);
      
      allocv (ndofe,r);
      allocv (ndofe,f);
      cn = new long [ndofe];
      
      Tt->give_code_numbers (i,cn);
      
      //  there must be this function (prescvalues)
      //  no other function may be here
      //  only initial values are required
      //  for more details see manual
      prescvalues (r.a,cn,ndofe);
      
      //  nodal values
      mxv (km,r,f);
      
      cmulv (-1.0,f);
      
      //  localization of nodal values
      locglob (rhs,f.a,cn,ndofe);
      
      destrm (km);  destrv (f);  destrv (r);
      delete [] cn;
    }
  }
}

/**
   function assembles right hand side of the problem
   
   @param lcid - load case id
   @param rhs - array containing right hand side
   @param n - number of components of array rhs
   
   12.3.2002
*/
void trfel_right_hand_side (long lcid,double *rhs,long n)
{
  double *av;
  av = new double [n];
  
  //  all components are set to zero
  nullv (rhs,n);
  
  //  function assembles the part of the right hand side vector
  //  caused by initial conditions and prescribed values
  assemble_init (rhs);
  
  //contribution of gradients for homogenization
  if(Tp->homogt == 1){
    nullv (av,n);
    assemble_gradients_contrib(av);
    addv (rhs,av,n);
  }

  switch (Tp->tmatt){
  case onemedium:{
    if (Tp->tprob==nonlinear_stationary_problem){
      nullv (av,n);
      copyv (rhs,av,n);
      
      nullv (rhs+(lcid*2+0)*n,n);
      nullv (rhs+(lcid*2+1)*n,n);
      
      Tb->lc[lcid*2+0].assemble (lcid*2+0,rhs+(lcid*2+0)*n);
      addv (rhs,av,n);
      //Tb->lc[lcid*2+1].assemble (lcid*2+1,rhs+(lcid*2+1)*n);
      //addv (rhs+(lcid*2+1)*n,av,n);
    }
    else{
      //  components of the vector av must be set to zero
      //  function assemble adds contributions to the input vector
      nullv (av,n);
      Tb->lc[lcid].assemble (lcid,av+lcid*n);
      addv (rhs,av,n);
    }
    break;
  }
  case twomediacoup:{
    //  components of the vector av must be set to zero
    //  function assemble adds contributions to the input vector
    nullv (av,n);
    Tb->lc[0].assemble (0,av);
    addv (rhs,av,n);
    
    //  components of the vector av must be set to zero
    //  function assemble adds contributions to the input vector
    nullv (av,n);
    Tb->lc[1].assemble (1,av);
    addv (rhs,av,n);
    break;
  }
  case threemediacoup:{
    //  components of the vector av must be set to zero
    //  function assemble adds contributions to the input vector
    nullv (av,n);
    Tb->lc[0].assemble (0,av);
    addv (rhs,av,n);

    //  components of the vector av must be set to zero
    //  function assemble adds contributions to the input vector
    nullv (av,n);
    Tb->lc[1].assemble (1,av);
    addv (rhs,av,n);

    //  components of the vector av must be set to zero
    //  function assemble adds contributions to the input vector
    nullv (av,n);
    Tb->lc[2].assemble (2,av);
    addv (rhs,av,n);
    break;
  }
  case fourmediacoup:{
    //  components of the vector av must be set to zero
    //  function assemble adds contributions to the input vector
    nullv (av,n);
    Tb->lc[0].assemble (0,av);
    addv (rhs,av,n);

    //  components of the vector av must be set to zero
    //  function assemble adds contributions to the input vector
    nullv (av,n);
    Tb->lc[1].assemble (1,av);
    addv (rhs,av,n);

    //  components of the vector av must be set to zero
    //  function assemble adds contributions to the input vector
    nullv (av,n);
    Tb->lc[2].assemble (2,av);
    addv (rhs,av,n);

    //  components of the vector av must be set to zero
    //  function assemble adds contributions to the input vector
    nullv (av,n);
    Tb->lc[3].assemble (3,av);
    addv (rhs,av,n);
    break;
  }
  default:{
    print_err("unknown number of transported matters is required",__FILE__,__LINE__,__func__);
  }
  }
  
  delete [] av;
}


/**
   function assembles flux on the boundary
   
   @param lcid - load case id
   @param iflux - array containing flux (right hand side)
   
   12.3.2002
*/
void trfel_bound_flux (long lcid,double *iflux,long n)
{
  double *av;
  av = new double [n];
  
  switch (Tp->tmatt){
  case onemedium:{
    Tb->lc[lcid].assemble_flux (lcid,av+lcid*n,n);
    addv (iflux,av,n);
    break;
  }
  case twomediacoup:{
    Tb->lc[0].assemble_flux (0,av,n);
    addv (iflux,av,n);
    nullv (av,n);
    Tb->lc[1].assemble_flux (1,av,n);
    addv (iflux,av,n);
    break;
  }
  case threemediacoup:{
    Tb->lc[0].assemble_flux (0,av,n);
    addv (iflux,av,n);
    nullv (av,n);
    Tb->lc[1].assemble_flux (1,av,n);
    addv (iflux,av,n);
    nullv (av,n);
    Tb->lc[2].assemble_flux (2,av,n);
    addv (iflux,av,n);
    break;
  }
  case fourmediacoup:{
    Tb->lc[0].assemble_flux (0,av,n);
    addv (iflux,av,n);
    nullv (av,n);
    Tb->lc[1].assemble_flux (1,av,n);
    addv (iflux,av,n);
    nullv (av,n);
    Tb->lc[2].assemble_flux (2,av,n);
    addv (iflux,av,n);
    nullv (av,n);
    Tb->lc[3].assemble_flux (3,av,n);
    addv (iflux,av,n);
    break;
  }
  default:{
    print_err("unknown number of transported matters is required",__FILE__,__LINE__,__func__);
  }
  }
  
  delete [] av;
}


/**
   function computes all required values
   this function is called e.g. before output print

   @param lcid - load case id

   1.4.2004, JK

   01/02/2010 changed by TKr
*/
void compute_req_valt (long lcid)
{
  // gradient computaion
  if (Tp->gradcomp == 1){
    //if (Tp->gradpos==1)
    //approximation();//it is already computed
    if (Tp->gradpos==2)
      compute_nodegrads ();
    //if (Tp->gradpos==3)//computed directly at nodes not implemented yet
    //compute_nodegrads ();
  }
  
  // flux computaion
  if (Tp->fluxcomp == 1){
    // gradients have to be computed if they are not computed yet
    //if (Tp->fluxpos==1)
    //approximation();//it is already computed
    if (Tp->fluxpos==1)
      compute_ipfluxes ();
    if (Tp->fluxpos==2)
      compute_nodefluxes ();
    //if (Tp->fluxpos==3)//computed directly at nodes not implemented yet
    //compute_nodefluxes ();
  }
    
  if (Tp->tprob == nonlinear_nonstationary_problem ||
      Tp->tprob == nonstationary_problem ||
      Tp->tprob == discont_nonstat_problem ||
      Tp->tprob == discont_nonlin_nonstat_problem ||
      Tp->tprob == growing_np_problem ||
      Tp->tprob == growing_np_problem_nonlin){

    // other computaion
    if (Tp->othercomp==1){
      if (Tp->otherpos==1)
	compute_ipotherst ();
      if (Tp->otherpos==2)
	compute_nodeotherst ();
      if (Tp->otherpos==3)//computed directly at nodes
	compute_nodeotherst_comp ();
    }
    
    // eqother computaion
    if (Tp->eqothercomp==1){
      if (Tp->eqotherpos==1)
	compute_ipeqotherst ();
      if (Tp->eqotherpos==2)
	compute_nodeeqotherst ();
      //if (Tp->eqotherpos==3)//computed directly at nodes not implemented yet
      //compute_nodeeqotherst ();
    }

  }
}



void give_nodal_humid (double *gv,long mnt)
{
  long i,cn,compother;
  double *r;
  
  r = new double [Tp->ntm];
  
  switch (Tp->tmatt){
  case onemedium:{
    med1 m1;

    for (i=0;i<mnt;i++){
      cn=Tt->give_dof (i,0);
      if (cn>0)  gv[i] = Lsrst->lhsi[cn-1]+Lsrst->lhs[cn-1];
      if (cn<0)  gv[i] = Tb->lc[0].pv[0-cn-1].getval();

      //gv[i] = m1.compute_othervalues(1,0,r);
      //if(gv[i] >= 1.0)
      //gv[i] = 1.0;
      
    }
    break;
    
    gv[i] = m1.compute_othervalues(1,0,r);
    if(gv[i] >= 1.0)
      gv[i] = 1.0;
  }
  case twomediacoup:{
    med2 m2;
    
    for (i=0;i<mnt;i++){
      cn=Tt->give_dof (i,0);
      if (cn>0)  r[0] = Lsrst->lhsi[cn-1]+Lsrst->lhs[cn-1];
      if (cn==0) r[0] = 0.0;
      if (cn<0)  r[0] = Tb->lc[0].pv[0-cn-1].getval();
      
      cn=Tt->give_dof (i,1);
      if (cn>0)  r[1] = Lsrst->lhsi[cn-1]+Lsrst->lhs[cn-1];
      if (cn==0) r[1] = 0.0;
      if (cn<0)  r[1] = Tb->lc[0].pv[0-cn-1].getval();
      
      compother = 1;
      if ((Tm->ip[0].tm) == kunzel)
	compother = 0;

      gv[i] = m2.compute_othervalues(compother,0,r);
      if(gv[i] >= 1.0)
	gv[i] = 1.0;
    }
    break;
  }
  case threemediacoup:{
    med3 m3;
    
    for (i=0;i<mnt;i++){
      cn=Tt->give_dof (i,0);
      if (cn>0)  r[0] = Lsrst->lhsi[cn-1]+Lsrst->lhs[cn-1];
      if (cn==0) r[0] = 0.0;
      if (cn<0)  r[0] = Tb->lc[0].pv[0-cn-1].getval();
      
      cn=Tt->give_dof (i,1);
      if (cn>0)  r[1] = Lsrst->lhsi[cn-1]+Lsrst->lhs[cn-1];
      if (cn==0) r[1] = 0.0;
      if (cn<0)  r[1] = Tb->lc[0].pv[0-cn-1].getval();
      
      cn=Tt->give_dof (i,2);
      if (cn>0)  r[2] = Lsrst->lhsi[cn-1]+Lsrst->lhs[cn-1];
      if (cn==0) r[2] = 0.0;
      if (cn<0)  r[2] = Tb->lc[0].pv[0-cn-1].getval();
      
      gv[i] = m3.compute_othervalues(3,0,r);
      if(gv[i] >= 1.0)
	gv[i] = 1.0;
    }
    break;
  }
  case fourmediacoup:{
    med4 m4;
    
    for (i=0;i<mnt;i++){
      cn=Tt->give_dof (i,0);
      if (cn>0)  r[0] = Lsrst->lhsi[cn-1]+Lsrst->lhs[cn-1];
      if (cn==0) r[0] = 0.0;
      if (cn<0)  r[0] = Tb->lc[0].pv[0-cn-1].getval();
      
      cn=Tt->give_dof (i,1);
      if (cn>0)  r[1] = Lsrst->lhsi[cn-1]+Lsrst->lhs[cn-1];
      if (cn==0) r[1] = 0.0;
      if (cn<0)  r[1] = Tb->lc[0].pv[0-cn-1].getval();
      
      cn=Tt->give_dof (i,2);
      if (cn>0)  r[2] = Lsrst->lhsi[cn-1]+Lsrst->lhs[cn-1];
      if (cn==0) r[2] = 0.0;
      if (cn<0)  r[2] = Tb->lc[0].pv[0-cn-1].getval();
      
      cn=Tt->give_dof (i,3);
      if (cn>0)  r[3] = Lsrst->lhsi[cn-1]+Lsrst->lhs[cn-1];
      if (cn==0) r[3] = 0.0;
      if (cn<0)  r[3] = Tb->lc[0].pv[0-cn-1].getval();
      
      gv[i] = m4.compute_othervalues(4,0,r);
      if(gv[i] >= 1.0)
	gv[i] = 1.0;
    }
    break;
  }
  default:{
    print_err("unknown number of transported matters is required",__FILE__,__LINE__,__func__);
  }
  }
  
  delete [] r;
}


/**
   function corrects solution of algebraic system of equations
   unacceptable trial values are replaced by the limit values
   
   14.7.2005
*/
void solution_correction ()
{
  long i,j,k,cn,ne,nne,ndofn,lcid,ipp;
  ivector nodes;
  vector nv;

  //  number of elements
  ne = Tt->ne;
  
  //  load case id, default value is 0
  lcid=0;
  
  //  loop over elements
  for (i=0;i<ne;i++){
    //  number nodes on element
    nne = Tt->give_nne (i);
    
    //  allocation
    allocv (nne,nodes);

    //  node numbers on element
    Tt->give_elemnodes (i,nodes);
    
    //  integration point id
    ipp=Tt->elements[i].ipp[0][0];
    
    //  extraction of nodal values    
    for (j=0;j<nne;j++){
      //  number of DOFs on node
      ndofn = Tt->give_ndofn (nodes[j]);
      //  allocation
      allocv (ndofn,nv);
      //  nodal values
      nodval (lcid,nv.a,nodes[j]);
      //  correction of values
      Tm->values_correction (nv,ipp);

      //  replacement of values in array containing the solution
      for (k=0;k<ndofn;k++){
	cn = Tt->give_dof (nodes[j],k);
	if (cn>0){
	  Lsrst->lhs[cn-1]=nv[k]-Lsrst->lhsi[cn-1];
	}
      }
      
      //
      destrv (nv);
    }
    
    destrv (nodes);
  }
  
}



/**
   function computes number of cycles
   the function is intended for computation of frozing cycles
   
   @param plhs - nodal values from previous time step
   @param pplhs - nodal values from time step i-2
   
   11.9.2007, JK
*/
void compute_cycles (double *plhs,double *pplhs)
{
  long i,j,lcid,ndofn,nne,ipp,nc;
  double *r,*pr,*ppr;
  ivector nod;
  
  //  only one load case can be used
  lcid=0;
  
  if (Tt->ncycl == NULL){
    Tt->ncycl = new long [Tt->nn];
    Tt->aux = new long[Tt->nn];
    
    for (i=0;i<Tt->nn;i++){
      Tt->ncycl[i]=0;
    }
  }
  
  for (i=0;i<Tt->nn;i++){
    Tt->aux[i]=0;
  }
  
  for (i=0;i<Tt->ne;i++){
    //  number of nodes on element
    nne = Tt->give_nne (i);
    allocv (nne,nod);
    //  element nodes
    Tt->give_elemnodes (i,nod);
    //  integration point id
    ipp = Tt->elements[i].ipp[0][0];
    
    for (j=0;j<nne;j++){
      if (Tt->aux[nod[j]]>0)
	continue;
      
      //  number of nodal DOF
      ndofn = Tt->give_ndofn (nod[j]);
      r = new double [ndofn];
      pr = new double [ndofn];
      ppr = new double [ndofn];
      //  nodal values
      nodval (lcid,r,nod[j]);
      prevnodval (lcid,pr,nod[j],plhs);
      prevnodval (lcid,ppr,nod[j],pplhs);
      //  cycle detection
      nc=Tm->cycle_detection (r,pr,ppr,ipp);
      
      Tt->ncycl[nod[j]]+=nc;
      Tt->aux[nod[j]]=1;
      
      delete [] r;  delete [] pr;  delete [] ppr;
    } 
    
    destrv (nod);
  }
}


/**
   function copies nodal values from %vector of nodal values (%vector of solution)
   to the array nodval defined on node
   
   @param lcid - load case id
   
   JK, 8.1.2008
*/
void copy_nodval (long lcid)
{
  long i,ndofn;
  double *r;
  
  //  loop over nodes
  for (i=0;i<Tt->nn;i++){
    //  number of DOFs on node
    ndofn = Tt->give_ndofn (i);
    //  array allocation
    r = new double [ndofn];
    //  nodal values
    nodval (lcid,r,i);
    //  save value to array nodval on nodes
    Tt->nodes[i].save_nodval (r);
    
    delete [] r;
  }
}



/**
   function assembles part of the %vector of right hand side from gradients of PUC
   
   @param rhs - right hand side
   
   19/08/2010, TKr
*/
void assemble_gradients_contrib (double *rhs)
{
  long i,ii,jj,k,ndofe,*cn,lcid=0,ntm,ncomp;
  vector grad,f;
  matrix lm;
  long ne = Tt->ne;
  

  for (i=0;i<ne;i++){
    
    if (Gtt->leso[i]==1){
      //  only elements switched on are processed
      
      //  number of DOFs on element
      ndofe=Tt->give_ndofe(i);
      ntm = Tp->ntm;
      ncomp = Tt->give_ncomp(i);
      allocm (ndofe,ntm*ncomp,lm);
      
      //  computation of element l_t matrix
      ltmat (i,lcid,lm);
      
      allocv (ntm*ncomp,grad);
      allocv (ndofe,f);
      cn = new long [ndofe];
      
      Tt->give_code_numbers (i,cn);
      
      //  constant gradients
      k = 0;
      for(ii=0;ii<ntm;ii++){
	for(jj=0;jj<ncomp;jj++){
	  grad[k]=Tb->lc[ii].mastergrad[jj];
	  k++;
	}
      }

      //  nodal values
      mxv (lm,grad,f);
      
      cmulv (-1.0,f);
      
      //  localization of nodal values
      locglob (rhs,f.a,cn,ndofe);
      
      destrm (lm);  destrv (f);  destrv (grad);
      delete [] cn;
    }
  }
}



/**
   function assembles l matrix and l_t (transposed) matrix of PUC
   
   @param lcid - load case id
*/
void assemble_l_matrix (matrix &lm, matrix &ltm)
{
  long i,ndofe,*rcn,*ccn,lcid=0,ntm,ncomp;
  matrix lkm,lkmt;
  
  //  loop over elements
  for (i=0;i<Tt->ne;i++){
    if (Gtt->leso[i]==1){
      //  only elements switched on are processed
      
      //  number of DOFs on element
      ndofe=Tt->give_ndofe(i);
      ntm = Tp->ntm;
      ncomp = Tt->give_ncomp(i);
      allocm (ndofe,ntm*ncomp,lkmt);
      allocm (ntm*ncomp,ndofe,lkm);
      
      //  computation of element l_t matrix
      ltmat (i,lcid,lkmt);

      //  computation of element l matrix
      lmat (i,lcid,lkm);

      rcn = new long [ndofe];
      ccn = new long [ntm*ncomp];
      
      Gtt->give_code_numbers (i,rcn);
      if(ncomp == 1){
	ccn[0] = 1;
	ccn[1] = 2;
      }
      if(ncomp == 2){
	ccn[0] = 1;
	ccn[1] = 2;
	ccn[2] = 3;
 	ccn[3] = 4;
      }
      if(ncomp == 3){
	ccn[0] = 1;
	ccn[1] = 2;
	ccn[2] = 3;
 	ccn[3] = 4;
	ccn[4] = 5;
 	ccn[5] = 6;
      }
      
      //  localization of l_t matrix
      mat_localize (ltm,lkmt,rcn,ccn);

      //  localization of l matrix
      mat_localize (lm,lkm,ccn,rcn);      
      
      destrm (lkm);
      destrm (lkmt);
      delete [] rcn;
      delete [] ccn;
    }
  }
}


/**
   function assembles conductivity matrix of PUC
   
   @param lcid - load case id
*/
void assemble_conductivity_matrix (matrix &km)
{
  long i,ndofe,*cn;
  matrix lm;
  
  //  loop over elements
  for (i=0;i<Tt->ne;i++){
    if (Gtt->leso[i]==1){
      //  number of DOFs on element
      ndofe = Gtt->give_ndofe (i);
      
      allocm (ndofe,ndofe,lm);
      //  conductivity matrix of one element
      conductmat (i,0,lm);
      
      cn = new long [ndofe];
      Gtt->give_code_numbers (i,cn);
      
      mat_localize (km,lm,cn,cn);
      
      destrm (lm);
      delete [] cn;
    }
  }
}

/**
   function assembles average conductivity (D) matrix of PUC
   
   @param lcid - load case id
*/
void assemble_average_d_matrix (matrix &km, double &area)
{
  long i,ii,jj,ntm,ncomp;
  double elemarea,invarea;
  matrix lm;
  
  area = 0.0;
  //  loop over elements
  for (i=0;i<Tt->ne;i++){
    elemarea = 0.0;
    if (Gtt->leso[i]==1){

      ntm = Tp->ntm;
      ncomp = Tt->give_ncomp(i);

      allocm (ncomp*ntm,ncomp*ntm,lm);

      //  conductivity D matrix of one element
      averdmat (i,elemarea,lm);
      
      for (ii=0;ii<ntm*ncomp;ii++){
	for (jj=0;jj<ntm*ncomp;jj++){
	  km[ii][jj]+=lm[ii][jj]*elemarea;
	}
      }
      destrm (lm);
      
      area += elemarea;
    }
  }

  invarea = 1.0/area;
  cmulm(invarea,km);

}

/**
   function assembles average capacity (C) matrix of PUC
   
   @param lcid - load case id
*/
void assemble_average_c_matrix (matrix &cm)
{
  long i,ii,jj,ntm,ncomp;
  double elemarea,area;
  matrix lm;
  
  area = 0.0;
  //  loop over elements
  for (i=0;i<Tt->ne;i++){
    elemarea = 0.0;
    if (Gtt->leso[i]==1){

      ntm = Tp->ntm;
      ncomp = Tt->give_ncomp(i);

      allocm (ntm,ntm,lm);

      //  conductivity D matrix of one element
      avercmat (i,elemarea,lm);
      
      for (ii=0;ii<ntm;ii++){
	for (jj=0;jj<ntm;jj++){
	  cm[ii][jj]+=lm[ii][jj]*elemarea/Tb->lc[jj].masterval;
	}
      }
      destrm (lm);
      
      area += elemarea;
    }
  }

  area = 1.0/area;
  cmulm(area,cm);

}
