/*
  File:             homogmat.cpp
  Author:           Jaroslav Kruis, 23.9.2010
  Purpose:          Calculates properties of homogenized material
*/ 

#include "homogmat.h"

homogmat::homogmat (void)
{

}
homogmat::~homogmat (void)
{
  destrm(dd);
  destrm(cc);
}




/**
   function computes conductivity %matrix of the material
   in the required integration point
   
   @param d - conductivity %matrix of material
   @param ri  - row index
   @param ci  - column index   
   @param ipp - number of integration point
   
   JK, 23.9.2010
*/
void homogmat::matcond (matrix &d,long ri,long ci,long ipp)
{
  long n;
  n = d.n;

  switch (n){
  case 1:{
    matcond1d (d,ri,ci,ipp);//1D
    break;
  }
  case 2:{
    matcond2d (d,ri,ci,ipp);//2D
    break;
  }
  case 3:{
    matcond3d (d,ri,ci,ipp);//3D
    break;
  }
  default:{
    print_err("unknown number of components of conductivity tensor is required",__FILE__,__LINE__,__func__);
  }
  }
}


/**
   function creates conductivity %matrix of the material for 1D problems

   @param d   - conductivity %matrix of the material
   @param ri  - row index
   @param ci  - column index   
   @param ipp - number of integration point
   
   JK, 23.9.2010
*/
void homogmat::matcond1d (matrix &d,long ri,long ci,long ipp)
{
  d[0][0] = dd[ri][ci];
}

/**
   function creates conductivity %matrix of the material for 2D problems

   @param d   - conductivity %matrix of the material
   @param ri  - row index
   @param ci  - column index   
   @param ipp - number of integration point
   
   JK, 23.9.2010
*/
void homogmat::matcond2d (matrix &d,long ri,long ci,long ipp)
{
  d[0][0] = dd[2*ri+0][2*ci+0];   d[0][1] = dd[2*ri+0][2*ci+1];
  d[1][0] = dd[2*ri+1][2*ci+0];   d[1][1] = dd[2*ri+1][2*ci+1];
}

/**
   function creates conductivity %matrix of the material for 3D problems

   @param d   - conductivity %matrix of the material
   @param ri  - row index
   @param ci  - column index   
   @param ipp - number of integration point
   
   JK, 23.9.2010
*/
void homogmat::matcond3d (matrix &d,long ri,long ci,long ipp)
{
  d[0][0] = dd[3*ri+0][3*ci+0];   d[0][1] = dd[3*ri+0][3*ci+1];   d[0][2] = dd[3*ri+0][3*ci+2];
  d[1][0] = dd[3*ri+1][3*ci+0];   d[1][1] = dd[3*ri+1][3*ci+1];   d[1][2] = dd[3*ri+1][3*ci+2];
  d[2][0] = dd[3*ri+2][3*ci+0];   d[2][1] = dd[3*ri+2][3*ci+1];   d[2][2] = dd[3*ri+2][3*ci+2];
}


/**
   function creates capacity %matrix of the material

   @param c   - capacity %matrix of the material
   @param ri  - row index
   @param ci  - column index   
   @param ipp - number of integration point
   
   JK, 23.9.2010
*/
void homogmat::matcap (double &c,long ri,long ci,long ipp)
{
  c = cc[ri][ci];
}

/**
   function assembles the conductivity %matrix after homogenization
   
   @param d - array of %matrix entries
   
   JK, 23.9.2010
*/
void homogmat::assemble_matrices (double *d)
{
  dd[0][0]=d[0];   dd[0][1]=d[1];   dd[0][2]=d[2];   dd[0][3]=d[3];
  dd[1][0]=d[4];   dd[1][1]=d[5];   dd[1][2]=d[6];   dd[1][3]=d[7];
  dd[2][0]=d[8];   dd[2][1]=d[9];   dd[2][2]=d[10];  dd[2][3]=d[11];
  dd[3][0]=d[12];  dd[3][1]=d[13];  dd[3][2]=d[14];  dd[3][3]=d[15];
  
  cc[0][0]=d[16];  cc[0][1]=0.0;
  cc[1][0]=0.0;    cc[1][1]=d[17];
}
