/*
  File:     lintett.cpp
  Author:   Jaroslav Kruis, 31.3.2003
  Purpose:  tetrahedral element with linear approximation functions
*/

#include "globalt.h"
#include "lintett.h"
#include "genfile.h"
#include "globmatt.h"
#include <math.h>

lintett::lintett (void)
{
  long i;

  //  number of nodes on element
  nne=4;
  //  number of edges
  ned=6;
  //  number of nodes on one edge
  nned=2;
  //  number of surfaces
  nsurf=4;
  //  number of nodes on one surface
  nnsurf=3;
  //  geometric problem dimension (3D)
  ncomp=3;

  intordb = 3;
  ntm=Tp->ntm;
  
  dofe = new long* [ntm];
  nip = new long* [ntm];
  intordkm = new long* [ntm];
  intordcm = new long* [ntm];
  ordering = new long* [ntm];
  for (i=0;i<ntm;i++){
    dofe[i] = new long [ntm];
    nip[i] = new long [ntm];
    intordkm[i] = new long [ntm];
    intordcm[i] = new long [ntm];
    ordering[i] = new long [nne];
  }

  switch (Tp->tmatt){
  case nomedium:{  break; }
  case onemedium:{
    ordering[0][0]=1;  ordering[0][1]=2;  ordering[0][2]=3;  ordering[0][3]=4; 
    
    dofe[0][0]=4;  intordkm[0][0]=1;  intordcm[0][0]=4;  nip[0][0]=5;
    ndofe=4;  napfun=1;
    break;
  }

  case twomediacoup:{
    ordering[0][0]=1;  ordering[0][1]=3;  ordering[0][2]=5;  ordering[0][3]=7;
    ordering[1][0]=2;  ordering[1][1]=4;  ordering[1][2]=6;  ordering[1][3]=8;

    intordkm[0][0]=1;  intordkm[0][1]=1;  intordkm[1][0]=1;  intordkm[1][1]=1;
    intordcm[0][0]=4;  intordcm[0][1]=4;  intordcm[1][0]=4;  intordcm[1][1]=4;
    nip[0][0]=5;       nip[0][1]=5;       nip[1][0]=5;       nip[1][1]=5;

    dofe[0][0]=4;  dofe[0][1]=4;  dofe[1][0]=4;  dofe[1][1]=4;
    ndofe=8;  napfun=2;
    break;
  }

  case threemediacoup:{
    ordering[0][0]=1;  ordering[0][1]=4;  ordering[0][2]=7;  ordering[0][3]=10;
    ordering[1][0]=2;  ordering[1][1]=5;  ordering[1][2]=8;  ordering[1][3]=11;
    ordering[2][0]=3;  ordering[2][1]=6;  ordering[2][2]=9;  ordering[2][3]=12;

    intordkm[0][0]=1;  intordkm[0][1]=1;  intordkm[0][2]=1;
    intordkm[1][0]=1;  intordkm[1][1]=1;  intordkm[1][2]=1;
    intordkm[2][0]=1;  intordkm[2][1]=1;  intordkm[2][2]=1;

    intordcm[0][0]=4;  intordcm[0][1]=4;  intordcm[0][2]=4;
    intordcm[1][0]=4;  intordcm[1][1]=4;  intordcm[1][2]=4;
    intordcm[2][0]=4;  intordcm[2][1]=4;  intordcm[2][2]=4;

    nip[0][0]=5;       nip[0][1]=5;       nip[0][2]=5;
    nip[1][0]=5;       nip[1][1]=5;       nip[1][2]=5;
    nip[2][0]=5;       nip[2][1]=5;       nip[2][2]=5;

    dofe[0][0]=4;      dofe[0][1]=4;      dofe[0][2]=4;
    dofe[1][0]=4;      dofe[1][1]=4;      dofe[1][2]=4;
    dofe[2][0]=4;      dofe[2][1]=4;      dofe[2][2]=4;

    ndofe=12;  napfun=3;
    break;
  }

  default:{
    fprintf (stderr,"\n\n unknown type of transported matter is required");
    fprintf (stderr,"\n in function lintett (file %s, line %d).\n",__FILE__,__LINE__);
  }
  }
}


lintett::~lintett (void)
{
  long i;
  
  for (i=0;i<ntm;i++){
    delete [] dofe[i];
    delete [] nip[i];
    delete [] intordkm[i];
    delete [] intordcm[i];
    delete [] ordering[i];
  }
  delete [] dofe;
  delete [] nip;
  delete [] intordkm;
  delete [] intordcm;
  delete [] ordering;
}

void lintett::codnum (long *cn,long ri)
{
  long i;
  for (i=0;i<nne;i++){
    cn[i]=ordering[ri][i];
  }
}

/**
   function approximates function defined by nodal values
   
   @param volcoord - volume coordinates
   @param nodval - %vector nodal values
   
   JK, 24.3.2002
*/
double lintett::approx (vector &volcoord,vector &nodval)
{
  double f;

  scprd (volcoord,nodval,f);
  
  return f;
}

/**
   function computes values in integration points from nodal values
   
   @param eid - element id
   
   JK, 24.3.2002
*/
void lintett::intpointval (long eid)
{
  long i,k,ii,jj,ipp;
  double val;
  ivector cn(ndofe);
  vector volcoord(4),r(ndofe),t(nne),gp1,gp2,gp3,w;
  
  Tt->give_code_numbers (eid,cn.a);
  elemvalues (0,eid,r.a,cn.a,ndofe);
  
  for (k=0;k<Tp->ntm;k++){
    
    for (i=0;i<dofe[k][k];i++){
      t[i]=r[ordering[k][i]-1];
    }
    
    for (ii=0;ii<Tp->ntm;ii++){
      for (jj=0;jj<Tp->ntm;jj++){
        
        allocv (intordkm[ii][jj],gp1);
        allocv (intordkm[ii][jj],gp2);
        allocv (intordkm[ii][jj],gp3);
        allocv (intordkm[ii][jj],w);
        gauss_points_tet (gp1.a,gp2.a,gp3.a,w.a,intordkm[ii][jj]);
        
        ipp=Tt->elements[eid].ipp[ii][jj];
        for (i=0;i<intordkm[ii][jj];i++){
          volcoord[0]=gp1[i];
          volcoord[1]=gp2[i];
          volcoord[2]=gp3[i];
          volcoord[3]=1.0-volcoord[0]-volcoord[1]-volcoord[2];
          
          val = approx (volcoord,t);
          Tm->ip[ipp].av[k]=val;
          ipp++;
        }
        
        destrv (gp1);  destrv (gp2);  destrv (gp3);  destrv (w);
        
        allocv (intordcm[ii][jj],gp1);
        allocv (intordcm[ii][jj],gp2);
        allocv (intordcm[ii][jj],gp3);
        allocv (intordcm[ii][jj],w);
        gauss_points_tet (gp1.a,gp2.a,gp3.a,w.a,intordcm[ii][jj]);
        for (i=0;i<intordcm[ii][jj];i++){
          volcoord[0]=gp1[i];
          volcoord[1]=gp2[i];
          volcoord[2]=gp3[i];
          volcoord[3]=1.0-volcoord[0]-volcoord[1]-volcoord[2];
          
          val = approx (volcoord,t);
          Tm->ip[ipp].av[k]=val;
          ipp++;
        }
        
        destrv (gp1);  destrv (gp2);  destrv (gp3);  destrv (w);
      }
    }
  }
}


/**
   function computes values in integration points from nodal values
   
   @param eid - element id
   
   JK, 24.3.2002
*/
void lintett::intpointgrad (long eid)
{
  long i,k,ii,jj,ipp;
  double det;
  ivector cn(ndofe);
  vector x(nne),y(nne),z(nne),volcoord(4),bb(4),cc(4),dd(4),r(ndofe),t(nne),gp1,gp2,gp3,w,grad(ncomp);
  matrix gm(ncomp,nne);

  Tt->give_node_coord3d (x,y,z,eid);
  Tt->give_code_numbers (eid,cn.a);
  elemvalues (0,eid,r.a,cn.a,ndofe);
  
  det = det3d (x.a,y.a,z.a);
  volb_3d (bb.a,y.a,z.a,det);
  volc_3d (cc.a,x.a,z.a,det);
  vold_3d (dd.a,x.a,y.a,det);
  
  //  matrix of gradients
  grad_matrix (gm,bb,cc,dd);

  for (k=0;k<Tp->ntm;k++){
    
    for (i=0;i<dofe[k][k];i++){
      t[i]=r[ordering[k][i]-1];
    }
    
    for (ii=0;ii<Tp->ntm;ii++){
      for (jj=0;jj<Tp->ntm;jj++){
        
        allocv (intordkm[ii][jj],gp1);
        allocv (intordkm[ii][jj],gp2);
        allocv (intordkm[ii][jj],gp3);
        allocv (intordkm[ii][jj],w);
        gauss_points_tet (gp1.a,gp2.a,gp3.a,w.a,intordkm[ii][jj]);
        
        ipp=Tt->elements[eid].ipp[ii][jj];
        for (i=0;i<intordkm[ii][jj];i++){
          volcoord[0]=gp1[i];
          volcoord[1]=gp2[i];
          volcoord[2]=gp3[i];
          volcoord[3]=1.0-volcoord[0]-volcoord[1]-volcoord[2];
          
          mxv (gm,t,grad);
          //Tm->storegrad (k,ipp,0,grad);
          Tm->storegrad (k,ipp,grad);
          ipp++;
        }
        
        destrv (gp1);  destrv (gp2);  destrv (gp3);  destrv (w);
        
        allocv (intordcm[ii][jj],gp1);
        allocv (intordcm[ii][jj],gp2);
        allocv (intordcm[ii][jj],gp3);
        allocv (intordcm[ii][jj],w);
        gauss_points_tet (gp1.a,gp2.a,gp3.a,w.a,intordcm[ii][jj]);
        for (i=0;i<intordcm[ii][jj];i++){
          volcoord[0]=gp1[i];
          volcoord[1]=gp2[i];
          volcoord[2]=gp3[i];
          volcoord[3]=1.0-volcoord[0]-volcoord[1]-volcoord[2];
          
          mxv (gm,t,grad);
          //Tm->storegrad (k,ipp,0,grad);
          Tm->storegrad (k,ipp,grad);
          ipp++;
        }
        
        destrv (gp1);  destrv (gp2);  destrv (gp3);  destrv (w);
      }
    }
  }
}

/**
   function approximates nodal values of array other to integration points
   
   @param eid - element id
   
   JK, 17.9.2005
*/
void lintett::intpointother (long eid)
{
  long i,k,ii,jj,ipp,ncompo;
  double val,*r;
  ivector nodes(nne);
  vector volcoord(4),t(nne),gp1,gp2,gp3,w;
  
  //  nodes of required element
  Tt->give_elemnodes (eid,nodes);
  
  //  number of components
  ncompo=Tt->nodes[i].ncompother;
  r = new double [ncompo*nne];
  
  //  nodal values of array other
  nodalotherval (r,nodes);
  
  for (k=0;k<ncompo;k++){
    
    for (i=0;i<nne;i++){
      t[i]=r[i*ncompo+k];
    }
    
    for (ii=0;ii<Tp->ntm;ii++){
      for (jj=0;jj<Tp->ntm;jj++){
        
        allocv (intordkm[ii][jj],gp1);
        allocv (intordkm[ii][jj],gp2);
        allocv (intordkm[ii][jj],gp3);
        allocv (intordkm[ii][jj],w);
        gauss_points_tet (gp1.a,gp2.a,gp3.a,w.a,intordkm[ii][jj]);
        
        ipp=Tt->elements[eid].ipp[ii][jj];
        for (i=0;i<intordkm[ii][jj];i++){
          volcoord[0]=gp1[i];
          volcoord[1]=gp2[i];
          volcoord[2]=gp3[i];
          volcoord[3]=1.0-volcoord[0]-volcoord[1]-volcoord[2];
          
          val = approx (volcoord,t);
          Tm->ip[ipp].other[k]=val;
          ipp++;
        }
        
        destrv (gp1);  destrv (gp2);  destrv (gp3);  destrv (w);
        
        allocv (intordcm[ii][jj],gp1);
        allocv (intordcm[ii][jj],gp2);
        allocv (intordcm[ii][jj],gp3);
        allocv (intordcm[ii][jj],w);
        gauss_points_tet (gp1.a,gp2.a,gp3.a,w.a,intordcm[ii][jj]);
        for (i=0;i<intordcm[ii][jj];i++){
          volcoord[0]=gp1[i];
          volcoord[1]=gp2[i];
          volcoord[2]=gp3[i];
          volcoord[3]=1.0-volcoord[0]-volcoord[1]-volcoord[2];
          
          val = approx (volcoord,t);
          Tm->ip[ipp].other[k]=val;
          ipp++;
        }
        
        destrv (gp1);  destrv (gp2);  destrv (gp3);  destrv (w);
      }
    }
  }
  delete [] r;
}

/**
   function assembles %matrix of base functions
   
   @param n - %matrix of base functions
   @param volcoord - %vector of volume coordinates
   
   JK, 24.3.2002
*/
void lintett::bf_matrix (matrix &n,vector &volcoord)
{
  fillm (0.0,n);
  
  n[0][0]=volcoord[0];
  n[0][1]=volcoord[1];
  n[0][2]=volcoord[2];
  n[0][3]=volcoord[3];
}

/**
   function assembles gradient of %matrix of base functions
   
   @param gm - gradient %matrix
   @param b,c,d - array containing coefficients of volume coordinates
   
   JK, 24.3.2002
*/
void lintett::grad_matrix (matrix &gm,vector &b,vector &c,vector &d)
{
  long i;
  fillm (0.0,gm);
  
  for (i=0;i<nne;i++){
    gm[0][i]=b[i];
    gm[1][i]=c[i];
    gm[2][i]=d[i];
  }
  
}

/**
   function computes conductivity %matrix of one transported medium
   
   @param lcid - load case id
   @param eid - element id
   @param ri, ci - row and column index
   @param km - conductivity %matrix
   
   JK, 24.3.2002
*/
void lintett::conductivity_matrix (long lcid,long eid,long ri,long ci,matrix &km)
{
  long ipp;
  double jac,det;
  vector x(nne),y(nne),z(nne),bb(4),cc(4),dd(4),volcoord(4),gp1,gp2,gp3,w;//tady??!!
  matrix gm(ncomp,dofe[ri][ci]),d(ncomp,ncomp);
  
  matrix n(1,dofe[ri][ci]);

  Tt->give_node_coord3d (x,y,z,eid);
  
  fillm (0.0,km);
  
  ipp=Tt->elements[eid].ipp[ri][ci];
  
  det = det3d (x.a,y.a,z.a);
  volb_3d (bb.a,y.a,z.a,det);
  volc_3d (cc.a,x.a,z.a,det);
  vold_3d (dd.a,x.a,y.a,det);
  
  allocv (intordkm[ri][ci],gp1);
  allocv (intordkm[ri][ci],gp2);
  allocv (intordkm[ri][ci],gp3);
  allocv (intordkm[ri][ci],w);

  gauss_points_tet (gp1.a,gp2.a,gp3.a,w.a,intordkm[ri][ci]);

  volcoord[0]=gp1[0];
  volcoord[1]=gp2[0];
  volcoord[2]=gp3[0];

  volcoord[3]=1.0-volcoord[0]-volcoord[1]-volcoord[2];

  //  matrix of gradients
  grad_matrix (gm,bb,cc,dd);
  
  //  matrix of conductivity of the material
  Tm->matcond (d,ipp,ri,ci);
  
  jac=fabs(det)/6.0;
  
  //  contribution to the conductivity matrix of the element
  bdbj (km.a,gm.a,d.a,jac,gm.m,gm.n);
  
  //convective terms
  if (Tp->tmatt == threemediacoup)
    {
      allocm(1,ncomp,d);
      
      Tm->matcond2(d,ipp,ri,ci);
      bf_matrix (n,volcoord);
      bdbjac(km, n, d, gm, jac);	
      destrm(d);
    }
  
  //coupled b.c.  //changed 10.8.2006
  if ((Tt->elements[eid].transi[lcid]==3) || (Tt->elements[eid].transi[lcid]==4)){
    transmission_matrix (lcid,eid,ri,ci,km);
  }
  
  destrv (gp1);
  destrv (gp2);
  destrv (gp3);
  destrv (w);
}


/**
   function computes capacity %matrix of one transported matter
   finite element with tri-linear approximation functions
   
   @param eid - number of element
   @param ri,ci - row and column indeces of the computed block in the resulting %matrix
   @param cm - capacity %matrix
   
   JK, 24.3.2002
*/
void lintett::capacity_matrix (long eid,long ri,long ci,matrix &cm)
{
  long i,ii;
  double jac,det,rho,c;
  ivector nodes(nne);
  vector x(nne),y(nne),z(nne),w(intordcm[ri][ci]),gp1(intordcm[ri][ci]),gp2(intordcm[ri][ci]),gp3(intordcm[ri][ci]);
  vector dens(nne),volcoord(4);
  matrix n(1,dofe[ri][ci]);
  
  Tt->give_elemnodes (eid,nodes);
  Tc->give_density (eid,nodes,dens);
  
  Tt->give_node_coord3d (x,y,z,eid);
  gauss_points_tet (gp1.a,gp2.a,gp3.a,w.a,intordcm[ri][ci]);
  det = det3d (x.a,y.a,z.a);
  
  fillm (0.0,cm);
  
  ii=Tt->elements[eid].ipp[ri][ci]+intordkm[ri][ci];

  for (i=0;i<intordcm[ri][ci];i++){
    volcoord[0]=gp1[i];
    volcoord[1]=gp2[i];
    volcoord[2]=gp3[i];
    volcoord[3]=1.0-volcoord[0]-volcoord[1]-volcoord[2];
    
    bf_matrix (n,volcoord);
    
    c=Tm->capcoeff (ii,ri,ci);
    rho = approx (volcoord,dens);
    jac=fabs(det)*w[i]*rho*c;
    
    nnj (cm.a,n.a,jac,n.m,n.n);
    ii++;

  }
}

/**
   function computes source vector of one matter on one element
   
   \int_{Omega} N^T N d Omega . s
   
   @param sv - source %vector of one matter
   @param lcid - load case id
   @param eid - element id
   @param ri,ci - row and column indeces of the block (ri must be equal to ci)
   
   JK, 24.3.2002
*/
void lintett::quantity_source_vector (vector &sv,vector &nodval,long eid,long ri,long ci)
{
  long i;
  double jac,det;
  vector x(nne),y(nne),z(nne),w(intordcm[ri][ci]),gp1(intordcm[ri][ci]),gp2(intordcm[ri][ci]),gp3(intordcm[ri][ci]),dens(nne);
  vector volcoord(4),v(dofe[ri][ci]);
  matrix n(1,dofe[ri][ci]),nm(dofe[ri][ci],dofe[ri][ci]);
  
  Tt->give_node_coord3d (x,y,z,eid);
  gauss_points_tet (gp1.a,gp2.a,gp3.a,w.a,intordcm[ri][ci]);
  det = fabs(det3d (x.a,y.a,z.a));
  
  fillm (0.0,nm);
  
  for (i=0;i<intordcm[ri][ci];i++){
    volcoord[0]=gp1[i];
    volcoord[1]=gp2[i];
    volcoord[2]=gp3[i];
    volcoord[3]=1.0-volcoord[0]-volcoord[1]-volcoord[2];
    
    bf_matrix (n,volcoord);
    
    jac=w[i]*det;
    
    nnj (nm.a,n.a,jac,n.m,n.n);
    
  }
  mxv (nm,nodval,v);  addv (sv,v,sv);
  
}


/**
   function computes internal fluxes
   
   @param lcid - load case id
   @param eid - element id
   @param ifl - %vector of internal fluxes
   
   JK, 3.12.2002
*/
void lintett::internal_fluxes (long lcid,long eid,vector &ifl)
{
  long ipp;
  double det,jac;
  ivector nodes(nne);
  vector x(nne),y(nne),z(nne),bb(4),cc(4),dd(4),fl(ncomp),contr(dofe[lcid][lcid]);
  matrix gm(ncomp,dofe[lcid][lcid]);
  
  Tt->give_elemnodes (eid,nodes);
  Tt->give_node_coord3d (x,y,z,eid);
  
  det = det3d (x.a,y.a,z.a);
  volb_3d (bb.a,y.a,z.a,det);
  volc_3d (cc.a,x.a,z.a,det);
  vold_3d (dd.a,x.a,y.a,det);
  
  ipp=Tt->elements[eid].ipp[lcid][lcid];
  
  Tm->computenlfluxes (lcid,ipp);
  
  //Tm->givefluxes (lcid,ipp,0,fl);
  Tm->givefluxes (lcid,ipp,fl);

  //  matrix of gradients
  grad_matrix (gm,bb,cc,dd);
  
  mtxv (gm,fl,contr);
  
  jac=fabs(det)/6.0;
  cmulv (jac,contr);

  addv (contr,ifl,ifl);
  
}

/**
   function assembles resulting element conductivity %matrix

   @param eid - element id
   @param lcid - load case id
   @param km - resulting conductivity %matrix of one element

   JK, 6.1.2002
*/
void lintett::res_conductivity_matrix (long eid,long lcid,matrix &km)
{
  long i,j,*rcn,*ccn;
  matrix lkm;
  
  for (i=0;i<ntm;i++){
    for (j=0;j<ntm;j++){
      rcn = new long [dofe[i][j]];
      ccn = new long [dofe[i][j]];
      allocm (dofe[i][j],dofe[i][j],lkm);
      conductivity_matrix (i,eid,i,j,lkm);
      codnum (rcn,i);
      codnum (ccn,j);
      mat_localize (km,lkm,rcn,ccn);
      destrm (lkm);
      delete [] rcn;  delete [] ccn;
    }
  }
}

/**
   function assembles resulting element capacity %matrix
   
   @param eid - element id
   @param cm - resulting capacity %matrix of one element

   JK, 6.1.2002
*/
void lintett::res_capacity_matrix (long eid,matrix &cm)
{
  long i,j,*rcn,*ccn,k,l;
  double s;
  matrix lcm;
  
  for (i=0;i<ntm;i++){
    for (j=0;j<ntm;j++){
      rcn = new long [dofe[i][j]];
      ccn = new long [dofe[i][j]];
      allocm (dofe[i][j],dofe[i][j],lcm);
      capacity_matrix (eid,i,j,lcm);

      ////////////////////////////////////////////////////
      // diagonalization of capacity matrix on one element
      if (Tp->diagcap == 1){
        for (k=0;k<dofe[i][j];k++){
          s=0.0;
          for (l=0;l<dofe[i][j];l++){
            s+=lcm[k][l];
            lcm[k][l]=0.0;
          }
          lcm[k][k]=s;
        }
      }
      ////////////////////////////////////////////////////

      codnum (rcn,i);
      codnum (ccn,j);
      mat_localize (cm,lcm,rcn,ccn);
      destrm (lcm);
      delete [] rcn;  delete [] ccn;
    }
  }
}

/**
   function assembles resulting element convection vector

   @param f - resulting convection %vector of one element
   @param lcid - load case id
   @param eid - element id
   @param leid - element id
   
   JK, 6.1.2002
*/
void lintett::res_convection_vector (vector &f,long lcid,long eid,long leid)
{
  long *cn;
  vector lf;

  cn = new long [dofe[lcid][lcid]];
  allocv (dofe[lcid][lcid],lf);
  convection_vector (lf,lcid,eid,leid,lcid,lcid);
  codnum (cn,lcid);
  locglob (f.a,lf.a,cn,dofe[lcid][lcid]);
  destrv (lf);
  delete [] cn;

  cmulv (-1.0,f,f);

}

/**
   function assembles resulting element transmission %vector

   @param f - resulting transmission %vector of one element
   @param lcid - load case id
   @param eid - element id
   @param leid - element id
   
   JK, 6.1.2002
*/
void lintett::res_transmission_vector (vector &f,long lcid,long eid,long leid)
{
  long *cn,i;
  vector lf;

  cn = new long [dofe[lcid][lcid]];
  codnum (cn,lcid);

  //coupled b.c.
  for (i=0;i<ntm;i++){    
    allocv (dofe[lcid][lcid],lf);
    
    if ((Tt->elements[eid].transi[i]==3) || (Tt->elements[eid].transi[i]==4)){
      transmission_vector (lf,i,eid,leid,lcid,i);
    }
    
    locglob (f.a,lf.a,cn,dofe[lcid][lcid]);
    destrv (lf);
       
  }
  
  delete [] cn;
}

/**
   function assembles resulting element source vector

   @param sv - resulting source %vector of one element
   @param lcid - load case id
   @param eid - element id
   
   JK, 6.1.2002
*/
void lintett::res_quantity_source_vector (vector &sv,vector &nodval,long lcid,long eid)
{
  long *cn;
  vector lsv;

  cn = new long [dofe[lcid][lcid]];
  allocv (dofe[lcid][lcid],lsv);
  quantity_source_vector (lsv,nodval,eid,lcid,lcid);
  codnum (cn,lcid);
  locglob (sv.a,lsv.a,cn,dofe[lcid][lcid]);
  destrv (lsv);
  delete [] cn;
}

/**
   function assembles resulting element internal fluxes vector

   @param eid - element id
   @param elemif - resulting internal fluxes %vector of one element
   
   JK, 6.1.2002
*/
void lintett::res_internal_fluxes (long eid,vector &elemif)
{
  long i,*cn;
  vector lif,tdnv;
  matrix cm;

  for (i=0;i<ntm;i++){
    cn = new long [dofe[i][i]];
    allocv (dofe[i][i],lif);
    internal_fluxes (i,eid,lif);
    codnum (cn,i);
    locglob (elemif.a,lif.a,cn,dofe[i][i]);

    destrv (lif);
    delete [] cn;
  }

  cn = new long [ndofe];
  Gtt->give_code_numbers (eid,cn);
  allocv (ndofe,tdnv);
  allocv (ndofe,lif);
  allocm (ndofe,ndofe,cm);
  nodalderivatives (tdnv.a,cn,ndofe);
  res_capacity_matrix (eid,cm);
  mxv (cm,tdnv,lif);

  subv (elemif,lif,elemif);
  
  delete [] cn;
  destrm (cm);
  destrv (lif);  destrv (tdnv);
}


/**
   function computes element quantity integral
   
   @param eid - element id
   @param nodval - %vector of quantity nodal values

   @retval f - element quantity integral

   TKr, 30.1.2004
*/
double lintett::total_integral(long eid,vector &nodval)
{
  long i;
  double jac,det,value,f;
  ivector nodes(nne);
  vector x(nne),y(nne),z(nne),w(4),gp1(4),gp2(4),gp3(4);
  vector dens(nne),volcoord(4);
  
  Tt->give_elemnodes (eid,nodes);
  
  Tt->give_node_coord3d (x,y,z,eid);
  gauss_points_tet (gp1.a,gp2.a,gp3.a,w.a,4);
  det = det3d (x.a,y.a,z.a);
  
  f = 0.0;
  
  for (i=0;i<4;i++){
    volcoord[0]=gp1[i];
    volcoord[1]=gp2[i];
    volcoord[2]=gp3[i];
    volcoord[3]=1.0-volcoord[0]-volcoord[1]-volcoord[2];
    
    value = approx (volcoord,nodval);
    jac=fabs(det)*w[i]*value;
    f = f + jac;
  }
  return(f);
}




/**
   function assembles resulting element boundary flux %vector

   @param f - resulting boundary flux %vector of one element
   @param lcid - load case id
   @param eid - element id
   @param leid - element id
   
   TKr, 28.2.2004
*/
void lintett::res_boundary_flux (vector &f,long lcid,long eid,long leid)
{
  long *cn,i;
  vector lf;

  cn = new long [dofe[lcid][lcid]];
  codnum (cn,lcid);
  
  //coupled b.c.
  for (i=0;i<ntm;i++){    
    allocv (dofe[lcid][lcid],lf);
    
    if ((Tt->elements[eid].transi[i]==3) || (Tt->elements[eid].transi[i]==4)){
      boundary_flux (lf,i,eid,leid,lcid,i);
    }
    
    locglob (f.a,lf.a,cn,dofe[lcid][lcid]);
    destrv (lf);
    
  }
  
  delete [] cn;
}


/**
   function computes other values in nodes of element

   @param lcid - load case id
   @param eid - element id
   @param ri - row index
   @param ci - column index

   10.5.2002
*/
void lintett::nod_others (long lcid,long eid,long ri,long ci)
{
  long i,j,k,ipp,ncompother,ii;
  vector other,h,r(ndofe);
  ivector nod(nne),cn(ndofe);
  
  Tt->give_elemnodes (eid,nod);  
  Tt->give_code_numbers (eid,cn.a);
  elemvalues (lcid,eid,r.a,cn.a,ndofe);
  ncompother = Tm->givencompother();
  allocv (ncompother,other);
  allocv (ntm,h);

  ii = 0;
  for (i=0;i<nne;i++){
    ipp=Tt->elements[eid].ipp[ri][ci];
    
    for (j=0;j<ntm;j++){
      h[j] = r[ii];
      ii++;
    }
    
    for(k=0;k<ncompother;k++){
      other[k] = Tm->givecompother(k,ipp,h.a);
    }
    
    //Tt->nodes[nod[i]].storeother (ncompother,other);
  }

  destrv (other);  destrv (cn);  destrv (nod);  destrv (r);
  destrv (h);
}







// This bellow added by TKr 18/03/2008
/**
   function computes contribution to the flux vector
   
   \int_{Gamma_2} N^T N dGamma * nodal_flux_values
   
   @param v - %vector of node fluxes
   @param lcid - load case id
   @param eid - element id
   @param leid - loaded element id
   @param ri,ci - row and column indices
   
   JK, 8.10.2001, TKr, 18.3.2008
*/
void lintett::convection_vector (vector &v,long lcid,long eid,long leid,long ri,long ci)
{
  long i,ipp,*bc;
  vector x(nne),y(nne),z(nne),w(intordb),gp1(intordb),gp2(intordb);
  vector list(3*nsurf),trc(3*nsurf),nodval(nne),av(nne),coef(nne);
  matrix km(nne,nne);
  
  fillv (0.0,v);
  
  //  node coordinates on element
  Tt->give_node_coord3d (x,y,z,eid);
  
  //  coordinates and weights of integration points
  gauss_points_tr (gp1.a,gp2.a,w.a,intordb);
  
  //  indicators of boundary conditions
  bc = new long [nsurf];
  Tb->lc[lcid].elemload[leid].give_bc (bc);
  
  //  auxiliary coefficients, necessary for function surface_integral
  for (i=0;i<3*nsurf;i++){
    trc[i]=1.0;
  }

  //  nodal values of exterior variable
  Tb->lc[lcid].elemload[leid].give_nodval (lcid,list);
  
  //  integration point number
  if (Tp->savemode==0)
    ipp=Tt->elements[eid].ipp[ri][ci];
  if (Tp->savemode==1)
    ipp=Tt->elements[eid].ipp[0][0];
  
  //  loop over surfaces
  for (i=0;i<nsurf;i++){
    
    if (bc[i]==2 || bc[i]==3 || bc[i]==4){
      //  nodal values on actual surface
      surfnodeval (i,nodval,list);
      
      fillm (0.0,km);

      surfnodeval (i,coef,trc);

      //  matrix obtained from integration over surface
      surface_integral (i,x,y,z,intordb,gp1,gp2,w,coef,km);
      
      mxv (km,nodval,av);
      
      addv (v,av,v);
    }
  }
  
  delete [] bc;
}

/**
   function computes transmission complement to the conductivity %matrix for one matter
   
   \int_{Gamma_3} N^T c_{tr} N dGamma
   
   @param lcid - number of load case
   @param eid - number of element
   @param ri,ci - row and column indeces
   @param km - part of the conductivity %matrix
   
   JK, 24.3.2002
   TKr, 30.1.2004 - new added
*/
void lintett::transmission_matrix (long lcid,long eid,long ri,long ci,matrix &km)
{
  long i,ipp,leid,*bc;
  vector x(nne),y(nne),z(nne),w(intordb),gp1(intordb),gp2(intordb);
  vector trc(3*nsurf),coeff(nne);
  
  //  node coordinates on element
  Tt->give_node_coord3d (x,y,z,eid);
  
  //  coordinates and weights of integration points
  gauss_points_tr (gp1.a,gp2.a,w.a,intordb);

  //  correspondence between element ordering and boundary element ordering
  for (i=0;i<Tb->lc[lcid].neb;i++){
    if (Tb->lc[lcid].elemload[i].eid==eid){
      leid=i;
      break;
    }
  }

  //  indicators of boundary conditions
  bc = new long [nsurf];
  Tb->lc[lcid].elemload[leid].give_bc (bc);
  //  transmission coefficients
  Tb->lc[lcid].elemload[leid].give_trc (lcid,trc);
  
  
  //  integration point number
  if (Tp->savemode==0)
    ipp=Tt->elements[eid].ipp[ri][ci];
  if (Tp->savemode==1)
    ipp=Tt->elements[eid].ipp[0][0];

  //  loop over surfaces
  for (i=0;i<nsurf;i++){
    
    if (bc[i]==4 || bc[i]>10){
      //  transformation of coefficients
      transf_coeff (i,coeff,trc,eid,ri,ci,ipp,bc);
      
      //  matrix obtained from integration over surface
      surface_integral (i,x,y,z,intordb,gp1,gp2,w,coeff,km);
    }
  }
  
  delete [] bc;
}

/**
   function computes contributions to the transmission vector
   
   \int_{Gamma_3} N^T c_{tr} N dGamma * nodal_external_temperature
   
   @param tmv - transmission %vector of one matter
   @param lcid - number of load case
   @param eid - number of element
   @param leid - number of loaded element
   @param ri,ci - row and column indices
      
   JK, 24.3.2002
   TKr, 30.1.2004 - new added
*/
void lintett::transmission_vector (vector &v,long lcid,long eid,long leid,long ri,long ci)
{
  long i,ipp,*bc;
  vector x(nne),y(nne),z(nne),w(intordb),gp1(intordb),gp2(intordb);
  vector list(3*nsurf),trc(3*nsurf),trr(3*nsurf),nodval(nne),av(nne),coef(nne);
  matrix km(nne,nne);
  
  fillv (0.0,v);
  
  //  node coordinates on element
  Tt->give_node_coord3d (x,y,z,eid);
  
  //  coordinates and weights of integration points
  gauss_points_tr (gp1.a,gp2.a,w.a,intordb);
  
  //  indicators of boundary conditions
  bc = new long [nsurf];
  Tb->lc[lcid].elemload[leid].give_bc (bc);
  //  nodal values of exterior variable
  Tb->lc[lcid].elemload[leid].give_nodval (lcid,list);
  //  transmission coefficients
  Tb->lc[lcid].elemload[leid].give_trc (lcid,trc);
  //  transmission/radiation coefficients
  Tb->lc[lcid].elemload[leid].give_trr (lcid,trr);

  //  integration point number
  if (Tp->savemode==0)
    ipp=Tt->elements[eid].ipp[ri][ci];
  if (Tp->savemode==1)
    ipp=Tt->elements[eid].ipp[0][0];
  
  for (i=0;i<nsurf;i++){
    
    if (bc[i]==4 || bc[i]>10){
      //  transformation of nodal values
      transf_val (i,nodval,list,trc,trr,eid,ri,ci,ipp,bc);
      //  determination of transmission coefficients on edge
      transf_coeff (i,coef,trc,eid,ri,ci,ipp,bc);

      fillm (0.0,km);
      //  matrix obtained from integration over surfaces
      surface_integral (i,x,y,z,intordb,gp1,gp2,w,coef,km);
      //  matrix-vector multiplication
      mxv (km,nodval,av);
      //  contribution added to element vector
      addv (v,av,v);
    }
  }
  
  delete [] bc;
}


/**
   function computes contributions to the boundary flux from transmission %vector
   
   \int_{Gamma_3} N^T c_{tr} N dGamma * nodal_external_temperature

   @param v - %vector of boundary fluxes
   @param lcid - load case id
   @param eid - element id
   @param leid - loaded element id
   @param ri,ci - row and column indices
   
   TKr, 28.2.2004
*/
void lintett::boundary_flux (vector &v,long lcid,long eid,long leid,long ri,long ci)
{
  long i,ipp,*bc;
  vector x(nne),y(nne),z(nne),w(intordb),gp1(intordb),gp2(intordb);
  vector list(3*nsurf),trc(3*nsurf),trr(3*nsurf),nodval(nne),av(nne),coef(nne);
  matrix km(nne,nne);
  
  fillv (0.0,v);
  
  //  node coordinates on element
  Tt->give_node_coord3d (x,y,z,eid);
  
  //  coordinates and weights of integration points
  gauss_points_tr (gp1.a,gp2.a,w.a,intordb);
  
  //  indicators of boundary conditions
  bc = new long [nsurf];
  Tb->lc[lcid].elemload[leid].give_bc (bc);
  //  nodal values of exterior variable
  Tb->lc[lcid].elemload[leid].give_nodval (lcid,list);
  //  transmission coefficients
  Tb->lc[lcid].elemload[leid].give_trc (lcid,trc);
  //  transmission/radiation coefficients
  Tb->lc[lcid].elemload[leid].give_trr (lcid,trr);

  //  integration point number
  if (Tp->savemode==0)
    ipp=Tt->elements[eid].ipp[ri][ci];
  if (Tp->savemode==1)
    ipp=Tt->elements[eid].ipp[0][0];
  
  for (i=0;i<nsurf;i++){
    
    if (bc[i]==4 || bc[i]>10){
      //  transformation of nodal values
      transf_flux (i,nodval,list,trc,trr,eid,ri,ci,ipp,bc);
      //  determination of transmission coefficients on edge
      transf_coeff (i,coef,trc,eid,ri,ci,ipp,bc);

      fillm (0.0,km);
      //  matrix obtained from integration over surfaces
      surface_integral (i,x,y,z,intordb,gp1,gp2,w,coef,km);
      //  matrix-vector multiplication
      mxv (km,nodval,av);
      //  contribution added to element vector
      addv (v,av,v);
    }
  }
  
  delete [] bc;
}


/**
   function integrates N^T c N over surface
   
   @param surf - surface id (number of surface)
   @param x,y,z - coordinates of element nodes
   @param intord - order of numerical integration
   @param gp1,gp2, w  - coordinates and weights of integration points
   @param coef - array of nodal values of coefficient
   @param km - output %matrix
   
   JK
*/
void lintett::surface_integral (long surf,vector &x,vector &y,vector &z,long intord,vector &gp1,vector &gp2,vector &w,
                                vector &coef,matrix &km)
{
  long i;
  double jac,ipval;
  vector volcoord(4);
  matrix n(1,nne);
  
  if (surf==0){
    for (i=0;i<intord;i++){
      volcoord[0]=0.0;
      volcoord[1]=gp1[i];
      volcoord[2]=gp2[i];
      volcoord[3]=1.0-volcoord[0]-volcoord[1]-volcoord[2];
      
      bf_matrix (n,volcoord);
      
      jac2d_3d (jac,x,y,z,gp1[i],gp2[i],0);
        
      ipval=approx (volcoord,coef);
      
      jac*=w[i]*ipval;
      nnj (km.a,n.a,jac,n.m,n.n);
    }
  }
  
  if (surf==1){
    for (i=0;i<intord;i++){
      volcoord[0]=gp1[i];
      volcoord[1]=0.0;
      volcoord[2]=gp2[i];
      volcoord[3]=1.0-volcoord[0]-volcoord[1]-volcoord[2];
      
      bf_matrix (n,volcoord);
      
      jac2d_3d (jac,x,y,z,gp1[i],gp2[i],1);
        
      ipval=approx (volcoord,coef);
      
      jac*=w[i]*ipval;
      nnj (km.a,n.a,jac,n.m,n.n);
    }
  }
  
  if (surf==2){
    for (i=0;i<intord;i++){
      volcoord[0]=gp1[i];
      volcoord[1]=gp2[i];
      volcoord[2]=0.0;
      volcoord[3]=1.0-volcoord[0]-volcoord[1]-volcoord[2];
      
      bf_matrix (n,volcoord);
      
      jac2d_3d (jac,x,y,z,gp1[i],gp2[i],2);
        
      ipval=approx (volcoord,coef);
      
      jac*=w[i]*ipval;
      nnj (km.a,n.a,jac,n.m,n.n);
    }
  }
  
  if (surf==3){
    for (i=0;i<intord;i++){
      volcoord[0]=gp1[i];
      volcoord[1]=gp2[i];
      volcoord[3]=0.0;
      volcoord[2]=1.0-volcoord[0]-volcoord[1]-volcoord[3];
      
      bf_matrix (n,volcoord);
      
      jac2d_3d (jac,x,y,z,gp1[i],gp2[i],3);
        
      ipval=approx (volcoord,coef);
      
      jac*=w[i]*ipval;
      nnj (km.a,n.a,jac,n.m,n.n);
    }
  }
  

}

void lintett::transf_flux (long surf,vector &coeff,vector &list,vector &trc,vector &trr,long eid,long ri,long ci,long ipp,long *bc)
{
  long i,j,k;
  double tr,new_flux;
  ivector nodes(nne),surfnod(nnsurf);

  //  zeroing of array of coefficients
  fillv (0.0,coeff);
  
  //  global numbers of nodes of element
  Tt->give_elemnodes (eid,nodes);

  //  local numbers of nodes on required surface
  lintetrahedral_surfnod (surfnod.a,surf);

  //  actual position in the array list
  j=surf*nnsurf;

  for (i=0;i<nnsurf;i++){
    if (bc[surf]==90){
      tr = trr[j+i]/trc[j+i];
    }

    //  node number
    k=surfnod[i];
    
    Tm->transmission_flux(new_flux,list[j+i],tr,ri,ci,nodes[k],bc[surf],ipp);
    
    coeff[k]=new_flux;
  }
  
}


void lintett::transf_coeff (long surf,vector &coeff,vector &list,long eid,long ri,long ci,long ipp,long *bc)
{
  long i,j,k;
  double new_coeff;
  ivector nodes(nne),surfnod(nnsurf);

  //  zeroing of array of coefficients
  fillv (0.0,coeff);
  
  //  global numbers of nodes of element
  Tt->give_elemnodes (eid,nodes);

  //  local numbers of nodes on required surface
  lintetrahedral_surfnod (surfnod.a,surf);

  //  actual position in the array list
  j=surf*nnsurf;

  for (i=0;i<nnsurf;i++){
    //  node number
    k=surfnod[i];
    
    Tm->transmission_transcoeff(new_coeff,list[j+i],ri,ci,nodes[k],bc[surf],ipp);
    
    coeff[k]=new_coeff;
  }
  
}


/**
   @param surf - number of required surface
   @param nodval - array of transformed nodal values
   @param list - array of nodal values defined on all surfaces
   @param trc -
   @param trr - 
   @param ri,ci - row and column indices
   @param ipp - integration point number
   @param bc - array defining boundary conditions
   
   JK, 19.8.2004
*/
void lintett::transf_val (long surf,vector &nodval,vector &list,vector &trc,vector &trr,long eid,long ri,long ci,long ipp,long *bc)
{
  long i,j,k;
  double tr,new_nodval;
  ivector nodes(nne),surfnod(nnsurf);
  
  //  zeroing of array of nodal values
  fillv (0.0,nodval);
  
  //  global numbers of nodes of element
  Tt->give_elemnodes (eid,nodes);

  //  local numbers of nodes on required surface
  lintetrahedral_surfnod (surfnod.a,surf);
  
  //  actual position in the array list
  j=surf*nnsurf;
  
  //  loop over number of nodes on one surface
  for (i=0;i<nnsurf;i++){
    tr = 0.0;
    if (bc[surf]==90){
      tr = trr[j+i]/trc[j+i];
    }
    
    //  node number
    k=surfnod[i];
    
    Tm->transmission_nodval(new_nodval,list[j+i],tr,ri,ci,nodes[k],bc[surf],ipp);
    
    //  transformed nodal values
    nodval[k]=new_nodval;
  }
}

/**
   function picks up nodal values on required surface
   
   @param surf - number of required surface
   @param nodval - array of nodal values
   @param list - array of nodal values defined on all surfaces
   
   JK, 19.8.2004
*/
void lintett::surfnodeval (long surf,vector &nodval,vector &list)
{
  long i,j;
  ivector surfnod(nnsurf);
  
  fillv (0.0,nodval);
  lintetrahedral_surfnod (surfnod.a,surf);
  
  for (i=0;i<nnsurf;i++){
    j=surfnod[i];
    nodval[j]=list[surf*nnsurf+i];
  }
}

